# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
from kaitaistruct import __version__ as ks_version, KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum


if parse_version(ks_version) < parse_version('0.7'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.7 or later is required, but you have %s" % (ks_version))

class Wfm1000z(KaitaiStruct):
    """Rigol DS1000Z scope .wmf file format.
    """

    class UnitEnum(Enum):
        w = 0
        a = 1
        v = 2
        u = 3

    class AcquistionEnum(Enum):
        normal = 0
        peak = 1
        average = 2
        high_resolution = 3

    class TimeModeEnum(Enum):
        yt = 0
        xy = 1
        roll = 2

    class ProbeEnum(Enum):
        x0_01 = 0
        x0_02 = 1
        x0_05 = 2
        x0_1 = 3
        x0_2 = 4
        x0_5 = 5
        x1 = 6
        x2 = 7
        x5 = 8
        x10 = 9
        x20 = 10
        x50 = 11
        x100 = 12
        x200 = 13
        x500 = 14
        x1000 = 15

    class BandwidthEnum(Enum):
        mhz_20 = 0
        no_limit = 1

    class CouplingEnum(Enum):
        dc = 0
        ac = 1
        gnd = 2
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        pass

    class ChannelSubhead(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.unknown_1 = self._io.read_bytes(3)
            self.unused_bits_1 = self._io.read_bits_int(7)
            self.enabled = self._io.read_bits_int(1) != 0
            self._io.align_to_byte()
            self.unknown_2 = self._io.read_bytes(7)
            self.unused_bits_2 = self._io.read_bits_int(7)
            self.inverted = self._io.read_bits_int(1) != 0
            self._io.align_to_byte()
            self.unknown_3 = self._io.read_bytes(10)
            self.probe_attenuation = self._io.read_s8le()
            self.unknown_4 = self._io.read_bytes(16)
            self.label = (KaitaiStream.bytes_terminate(self._io.read_bytes(4), 0, False)).decode(u"UTF-8")
            self.unknown_5 = self._io.read_bytes(10)


    class RawData(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            if self._root.header.stride == 1:
                self.raw1 = [None] * (self._root.header.points)
                for i in range(self._root.header.points):
                    self.raw1[i] = self._io.read_u1()


            if self._root.header.stride == 2:
                self.raw2 = [None] * (self._root.header.points)
                for i in range(self._root.header.points):
                    self.raw2[i] = self._io.read_u2le()


            if self._root.header.stride == 4:
                self.raw4 = [None] * (self._root.header.points)
                for i in range(self._root.header.points):
                    self.raw4[i] = self._io.read_u4le()




    class ChannelHead(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.enabled_val = self._io.read_u1()
            self.coupling = self._root.CouplingEnum(self._io.read_u1())
            self.bandwidth_limit = self._root.BandwidthEnum(self._io.read_u1())
            self.probe_type = self._io.read_u1()
            self.probe_ratio = self._root.ProbeEnum(self._io.read_u1())
            self.unused = self._io.read_bytes(3)
            self.scale = self._io.read_f4le()
            self.shift = self._io.read_f4le()
            self.inverted_val = self._io.read_u1()
            self.unit = self._root.UnitEnum(self._io.read_u1())
            self.unknown_2 = self._io.read_bytes(10)

        @property
        def y_scale(self):
            if hasattr(self, '_m_y_scale'):
                return self._m_y_scale if hasattr(self, '_m_y_scale') else None

            self._m_y_scale = (-(self.volt_per_division) / 20.0)
            return self._m_y_scale if hasattr(self, '_m_y_scale') else None

        @property
        def y_offset(self):
            if hasattr(self, '_m_y_offset'):
                return self._m_y_offset if hasattr(self, '_m_y_offset') else None

            self._m_y_offset = (self.shift - self.volt_per_division)
            return self._m_y_offset if hasattr(self, '_m_y_offset') else None

        @property
        def time_offset(self):
            if hasattr(self, '_m_time_offset'):
                return self._m_time_offset if hasattr(self, '_m_time_offset') else None

            self._m_time_offset = self._root.header.time_offset
            return self._m_time_offset if hasattr(self, '_m_time_offset') else None

        @property
        def inverted(self):
            if hasattr(self, '_m_inverted'):
                return self._m_inverted if hasattr(self, '_m_inverted') else None

            self._m_inverted = (True if self.inverted_val != 0 else False)
            return self._m_inverted if hasattr(self, '_m_inverted') else None

        @property
        def time_scale(self):
            if hasattr(self, '_m_time_scale'):
                return self._m_time_scale if hasattr(self, '_m_time_scale') else None

            self._m_time_scale = self._root.header.time_scale
            return self._m_time_scale if hasattr(self, '_m_time_scale') else None

        @property
        def volt_offset(self):
            if hasattr(self, '_m_volt_offset'):
                return self._m_volt_offset if hasattr(self, '_m_volt_offset') else None

            self._m_volt_offset = self.shift
            return self._m_volt_offset if hasattr(self, '_m_volt_offset') else None

        @property
        def volt_per_division(self):
            if hasattr(self, '_m_volt_per_division'):
                return self._m_volt_per_division if hasattr(self, '_m_volt_per_division') else None

            self._m_volt_per_division = ((-1.0 * self.scale) if self.inverted else (1.0 * self.scale))
            return self._m_volt_per_division if hasattr(self, '_m_volt_per_division') else None

        @property
        def volt_scale(self):
            if hasattr(self, '_m_volt_scale'):
                return self._m_volt_scale if hasattr(self, '_m_volt_scale') else None

            self._m_volt_scale = (self.volt_per_division / 25.0)
            return self._m_volt_scale if hasattr(self, '_m_volt_scale') else None

        @property
        def probe_value(self):
            if hasattr(self, '_m_probe_value'):
                return self._m_probe_value if hasattr(self, '_m_probe_value') else None

            self._m_probe_value = (0.01 if self.probe_ratio == self._root.ProbeEnum.x0_01 else (0.02 if self.probe_ratio == self._root.ProbeEnum.x0_02 else (0.05 if self.probe_ratio == self._root.ProbeEnum.x0_05 else (0.1 if self.probe_ratio == self._root.ProbeEnum.x0_1 else (0.2 if self.probe_ratio == self._root.ProbeEnum.x0_2 else (0.5 if self.probe_ratio == self._root.ProbeEnum.x0_5 else (1.0 if self.probe_ratio == self._root.ProbeEnum.x1 else (2.0 if self.probe_ratio == self._root.ProbeEnum.x2 else (5.0 if self.probe_ratio == self._root.ProbeEnum.x5 else (10.0 if self.probe_ratio == self._root.ProbeEnum.x10 else (20.0 if self.probe_ratio == self._root.ProbeEnum.x20 else (50.0 if self.probe_ratio == self._root.ProbeEnum.x50 else (100.0 if self.probe_ratio == self._root.ProbeEnum.x100 else (200.0 if self.probe_ratio == self._root.ProbeEnum.x200 else (500.0 if self.probe_ratio == self._root.ProbeEnum.x500 else 1000.0)))))))))))))))
            return self._m_probe_value if hasattr(self, '_m_probe_value') else None

        @property
        def enabled(self):
            if hasattr(self, '_m_enabled'):
                return self._m_enabled if hasattr(self, '_m_enabled') else None

            self._m_enabled = (True if self.enabled_val != 0 else False)
            return self._m_enabled if hasattr(self, '_m_enabled') else None


    class FileHeader(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.magic = self._io.ensure_fixed_contents(b"\x01\xFF\xFF\xFF")
            self.magic2 = self._io.read_u2le()
            self.structure_size = self._io.read_u2le()
            self.model_number = (KaitaiStream.bytes_terminate(self._io.read_bytes(20), 0, False)).decode(u"ascii")
            self.firmware_version = (KaitaiStream.bytes_terminate(self._io.read_bytes(20), 0, False)).decode(u"ascii")
            self.block = self._io.ensure_fixed_contents(b"\x01\x00")
            self.file_version = self._io.read_u2le()


    class WfmHeader(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.picoseconds_per_division = self._io.read_u8le()
            self.picoseconds_offset = self._io.read_s8le()
            self.crc = self._io.read_u4le()
            self.structure_size = self._io.ensure_fixed_contents(b"\xD8\x00")
            self.structure_version = self._io.read_u2le()
            self.unused_bits_1 = self._io.read_bits_int(4)
            self.ch4_enabled = self._io.read_bits_int(1) != 0
            self.ch3_enabled = self._io.read_bits_int(1) != 0
            self.ch2_enabled = self._io.read_bits_int(1) != 0
            self.ch1_enabled = self._io.read_bits_int(1) != 0
            self._io.align_to_byte()
            self.unused_mask_bytes = self._io.read_bytes(3)
            self.ch1_file_offset = self._io.read_u4le()
            self.ch2_file_offset = self._io.read_u4le()
            self.ch3_file_offset = self._io.read_u4le()
            self.ch4_file_offset = self._io.read_u4le()
            self.la_offset = self._io.read_u4le()
            self.acq_mode = self._root.AcquistionEnum(self._io.read_u1())
            self.average_time = self._io.read_u1()
            self.sample_mode = self._io.ensure_fixed_contents(b"\x00")
            self.time_mode = self._root.TimeModeEnum(self._io.read_u1())
            self.memory_depth = self._io.read_u4le()
            self.sample_rate_ghz = self._io.read_f4le()
            self.ch = [None] * (4)
            for i in range(4):
                self.ch[i] = self._root.ChannelHead(self._io, self, self._root)

            self.la_parameters = self._io.read_bytes(12)
            self.setup_size = self._io.read_u4le()
            self.setup_offset = self._io.read_u4le()
            self.horizontal_size = self._io.read_u4le()
            self.horizontal_offset = self._io.read_u4le()

        @property
        def points(self):
            if hasattr(self, '_m_points'):
                return self._m_points if hasattr(self, '_m_points') else None

            self._m_points = self.memory_depth // self.stride
            return self._m_points if hasattr(self, '_m_points') else None

        @property
        def ch2_int(self):
            if hasattr(self, '_m_ch2_int'):
                return self._m_ch2_int if hasattr(self, '_m_ch2_int') else None

            self._m_ch2_int = (1 if self.ch2_enabled else 0)
            return self._m_ch2_int if hasattr(self, '_m_ch2_int') else None

        @property
        def sample_rate_hz(self):
            if hasattr(self, '_m_sample_rate_hz'):
                return self._m_sample_rate_hz if hasattr(self, '_m_sample_rate_hz') else None

            self._m_sample_rate_hz = (self.sample_rate_ghz * 1E+9)
            return self._m_sample_rate_hz if hasattr(self, '_m_sample_rate_hz') else None

        @property
        def time_offset(self):
            if hasattr(self, '_m_time_offset'):
                return self._m_time_offset if hasattr(self, '_m_time_offset') else None

            self._m_time_offset = (self.picoseconds_offset * 1E-12)
            return self._m_time_offset if hasattr(self, '_m_time_offset') else None

        @property
        def ch4_int(self):
            if hasattr(self, '_m_ch4_int'):
                return self._m_ch4_int if hasattr(self, '_m_ch4_int') else None

            self._m_ch4_int = (1 if self.ch4_enabled else 0)
            return self._m_ch4_int if hasattr(self, '_m_ch4_int') else None

        @property
        def ch3_int(self):
            if hasattr(self, '_m_ch3_int'):
                return self._m_ch3_int if hasattr(self, '_m_ch3_int') else None

            self._m_ch3_int = (1 if self.ch3_enabled else 0)
            return self._m_ch3_int if hasattr(self, '_m_ch3_int') else None

        @property
        def total_channels(self):
            if hasattr(self, '_m_total_channels'):
                return self._m_total_channels if hasattr(self, '_m_total_channels') else None

            self._m_total_channels = (((self.ch1_int + self.ch2_int) + self.ch3_int) + self.ch4_int)
            return self._m_total_channels if hasattr(self, '_m_total_channels') else None

        @property
        def time_scale(self):
            if hasattr(self, '_m_time_scale'):
                return self._m_time_scale if hasattr(self, '_m_time_scale') else None

            self._m_time_scale = (self.picoseconds_per_division * 1E-12)
            return self._m_time_scale if hasattr(self, '_m_time_scale') else None

        @property
        def seconds_per_point(self):
            if hasattr(self, '_m_seconds_per_point'):
                return self._m_seconds_per_point if hasattr(self, '_m_seconds_per_point') else None

            self._m_seconds_per_point = (1 / self.sample_rate_hz)
            return self._m_seconds_per_point if hasattr(self, '_m_seconds_per_point') else None

        @property
        def ch1_int(self):
            if hasattr(self, '_m_ch1_int'):
                return self._m_ch1_int if hasattr(self, '_m_ch1_int') else None

            self._m_ch1_int = (1 if self.ch1_enabled else 0)
            return self._m_ch1_int if hasattr(self, '_m_ch1_int') else None

        @property
        def stride(self):
            if hasattr(self, '_m_stride'):
                return self._m_stride if hasattr(self, '_m_stride') else None

            self._m_stride = (4 if self.total_channels == 3 else self.total_channels)
            return self._m_stride if hasattr(self, '_m_stride') else None


    @property
    def preheader(self):
        if hasattr(self, '_m_preheader'):
            return self._m_preheader if hasattr(self, '_m_preheader') else None

        _pos = self._io.pos()
        self._io.seek(0)
        self._m_preheader = self._root.FileHeader(self._io, self, self._root)
        self._io.seek(_pos)
        return self._m_preheader if hasattr(self, '_m_preheader') else None

    @property
    def header(self):
        if hasattr(self, '_m_header'):
            return self._m_header if hasattr(self, '_m_header') else None

        _pos = self._io.pos()
        self._io.seek(64)
        self._m_header = self._root.WfmHeader(self._io, self, self._root)
        self._io.seek(_pos)
        return self._m_header if hasattr(self, '_m_header') else None

    @property
    def data(self):
        if hasattr(self, '_m_data'):
            return self._m_data if hasattr(self, '_m_data') else None

        _pos = self._io.pos()
        self._io.seek(((304 + self._root.header.setup_size) + self._root.header.horizontal_size))
        self._m_data = self._root.RawData(self._io, self, self._root)
        self._io.seek(_pos)
        return self._m_data if hasattr(self, '_m_data') else None





#pylint: disable=invalid-name
#pylint: disable=too-many-instance-attributes
#pylint: disable=too-many-return-statements
"""
Class structure and methods for an oscilloscope channel.

The idea is to collect all the relevant information from all the Rigol
scope waveforms into a single structure that can be handled in a uniform
and consistent manner.

Specifically this lets one just use

    channel.times   : numpy array of signal times
    channel.volts   : numpy array of signal voltages

or the stringification method to describe a channel

    print(channel)

"""
from enum import Enum
import numpy as np

class UnitEnum(Enum):
    """Enumerated units for scopes without them."""

    w = 0
    a = 1
    v = 2
    u = 3

def best_scale(number):
    """Scale and units for a number with proper prefix."""
    absnr = abs(number)

    if absnr == 0:
        return 1, ' '
    if absnr < 0.99999999e-9:
        return 1e12, 'p'
    if absnr < 0.99999999e-6:
        return 1e9, 'n'
    if absnr < 0.99999999e-3:
        return 1e6, 'µ'
    if absnr < 0.99999999:
        return 1e3, 'm'
    if absnr < 0.99999999e3:
        return 1, ' '
    if absnr < 0.99999999e6:
        return 1e-3, 'k'
    if absnr < 0.999999991e9:
        return 1e-6, 'M'
    return 1e-9, 'G'

def engineering_string(number, n_digits):
    """Format number with proper prefix."""
    scale, prefix = best_scale(number)
    fformat = "%%.%df %%s" % n_digits
    s = fformat % (number * scale, prefix)
    return s


### maybe replace with something like
#    x0[0::2] = data[offset_1:offset_1+pagesize]
#    x0[1::2] = data[offset_2:offset_2+pagesize]

def _channel_bytes(enabled_count, data, stride):
    """
    Return right series of bytes for a channel for 1000Z scopes.

    Waveform points are interleaved stored in memory when two or more
    channels are enabled:

    Only one channel enabled:
        CH1CH1CH1CH1

    Two channels Enabled:
        CH2CH1CH2CH1

    Three or Four channels enabled:
        CH4CH3CH2CH1

    Args:
        enabled_count: the number of enabled channels before this one
        data:          object containing the raw data structures
    Returns
        byte array for a particular channel
    """
    if stride == 1:
        raw_bytes = np.array(data.raw1, dtype=np.uint8)

    if stride == 2:
        if enabled_count == 0:
            raw_bytes = np.array(np.uint16(data.raw2) & 0x00FF, dtype=np.uint8)
        else:
            raw_bytes = np.array((np.uint16(data.raw2) & 0xFF00) >> 8, dtype=np.uint8)

    if stride == 4:
        if enabled_count == 3:
            raw_bytes = np.array(np.uint32(data.raw4)
                                 & 0x000000FF, dtype=np.uint8)
        elif enabled_count == 2:
            raw_bytes = np.array(
                (np.uint32(data.raw4) & 0x0000FF00) >> 8, dtype=np.uint8)
        elif enabled_count == 1:
            raw_bytes = np.array(
                (np.uint32(data.raw4) & 0x00FF0000) >> 16, dtype=np.uint8)
        else:
            raw_bytes = np.array(
                (np.uint32(data.raw4) & 0xFF000000) >> 24, dtype=np.uint8)

    return raw_bytes


class Channel():
    """Base class for a single channel."""

    def __init__(self, w, ch, scope, prior):
        """
        Initialize a Channel Object.

        Args:
            w: Wfm object
            ch: 1,2,3,4
            scope: string describing scope
            prior: number of channels that came before
        Returns:
            Channel object
        """
        self.channel_number = ch
        self.name = "CH %d" % ch
        self.waveform = w
        self.seconds_per_point = w.header.seconds_per_point
        self.firmware = 'unknown'
        self.unit = 'V'
        self.points = 0
        self.raw = None
        self.volts = None
        self.times = None
        self.coupling = 'unknown'
        self.roll_stop = 0
        self.time_offset = 0
        self.time_scale = 1

        if ch <= len(w.header.ch):
            channel = w.header.ch[ch-1]
            self.enabled = channel.enabled
            self.volt_scale = channel.volt_scale
            self.volt_offset = channel.volt_offset
            self.y_scale = channel.volt_scale
            self.y_offset = channel.volt_offset
            self.volt_per_division = channel.volt_per_division
            self.probe_value = channel.probe_value
            self.unit = channel.unit
            self.inverted = channel.inverted
        else:
            self.enabled = False
            self.volt_scale = 1
            self.volt_offset = 0
            self.y_scale = 1
            self.y_offset = 0
            self.volt_per_division = 1
            self.probe_value = 1
            self.unit = UnitEnum.v
            self.inverted = False

        if scope == 'wfm1000c':
            self.ds1000c(w, ch)
        elif scope == 'wfm1000e':
            self.ds1000e(w, ch)
        elif scope == 'wfm1000z':
            self.ds1000z(w, ch, prior)
        elif scope == 'wfm2000':
            self.ds2000(w, ch)
        elif scope == 'wfm4000':
            self.ds4000(w, ch)
        elif scope == 'wfm6000':
            self.ds6000(w, ch)


    def __str__(self):
        """Describe this channel."""
        s = "     Channel %d:\n" % self.channel_number
        s += "         Coupling = %8s\n" % self.coupling.rjust(7, ' ')
        s += "            Scale = %10sV/div\n" % engineering_string(self.volt_per_division, 2)
        s += "           Offset = %10sV\n" % engineering_string(self.volt_offset, 2)
        s += "            Probe = %7gX\n" % self.probe_value
        s += "         Inverted = %8s\n\n" % self.inverted
        s += "        Time Base = %10ss/div\n" % engineering_string(self.time_scale, 3)
        s += "           Offset = %10ss\n" % engineering_string(self.time_offset, 3)
        s += "            Delta = %10ss/point\n" % engineering_string(self.seconds_per_point, 3)
        s += "           Points = %8d\n\n" % self.points
        if self.enabled:
            s += "         Count    = [%9d,%9d,%9d  ... %9d,%9d]\n" % (
                1, 2, 3, self.points-1, self.points)
            s += "           Raw    = [%9d,%9d,%9d  ... %9d,%9d]\n" % (
                self.raw[0], self.raw[1], self.raw[2], self.raw[-2], self.raw[-1])
            t = [engineering_string(self.times[i], 3) +
                 "s" for i in [0, 1, 2, -2, -1]]
            s += "           Times  = [%9s,%9s,%9s  ... %9s,%9s]\n" % (
                t[0], t[1], t[2], t[-2], t[-1])
            v = [engineering_string(self.volts[i], 2) +
                 "V" for i in [0, 1, 2, -2, -1]]
            s += "           Volts  = [%9s,%9s,%9s  ... %9s,%9s]\n" % (
                v[0], v[1], v[2], v[-2], v[-1])
        return s

    def calc_times_and_volts(self):
        """Calculate the times and voltages for this channel."""
        if self.enabled:
            self.volts = self.y_scale * (127.0 - self.raw) - self.y_offset
            h = self.points * self.seconds_per_point / 2
            self.times = np.linspace(-h, h, self.points) + self.time_offset


    def ds1000c(self, w, ch):
        """Interpret waveform data for 1000CD series scopes."""
        if ch == 1:
            self.time_offset = w.header.ch1_time_offset
            self.time_scale = w.header.ch1_time_scale
            if self.enabled:
                self.points = len(w.data.ch1)
                self.raw = np.array(w.data.ch1, dtype=np.uint8)

        if ch == 2:
            self.time_offset = w.header.ch2_time_offset
            self.time_scale = w.header.ch2_time_scale
            if self.enabled:
                self.points = len(w.data.ch2)
                self.raw = np.array(w.data.ch2, dtype=np.uint8)

        self.calc_times_and_volts()


    def ds1000e(self, w, ch):
        """Interpret waveform data for 1000D and 1000E series scopes."""
        self.roll_stop = w.header.roll_stop

        if ch == 1:
            self.time_offset = w.header.ch1_time_offset
            self.time_scale = w.header.ch1_time_scale
            if self.enabled:
                self.points = len(w.data.ch1)
                self.raw = np.array(w.data.ch1, dtype=np.uint8)

        elif ch == 2:
            self.time_offset = w.header.ch2_time_offset
            self.time_scale = w.header.ch2_time_scale
            if self.enabled:
                self.points = len(w.data.ch2)
                self.raw = np.array(w.data.ch2, dtype=np.uint8)

        self.calc_times_and_volts()
#        if self.enabled:
#            self.volts *= self.probe_value

    def ds1000z(self, w, ch, enabled_count):
        """Interpret waveform for the Rigol DS1000Z series."""
        self.time_scale = w.header.time_scale
        self.time_offset = w.header.time_offset
        self.points = w.header.points
        self.stride = w.header.stride
        self.firmware = w.preheader.firmware_version
        self.probe = w.header.ch[ch-1].probe_value
        self.coupling = w.header.ch[ch-1].coupling.name.upper()
        self.y_scale = w.header.ch[ch-1].y_scale
        self.y_offset = w.header.ch[ch-1].y_offset

        if self.enabled:
            self.raw = _channel_bytes(enabled_count, w.data, self.stride)

        self.calc_times_and_volts()

    def ds2000(self, w, ch):
        """Interpret waveform for the Rigol DS2000 series."""
        self.time_offset = w.header.time_offset
        self.time_scale = w.header.time_scale
        self.points = w.header.storage_depth
        self.firmware = w.header.firmware_version
        self.unit = UnitEnum(w.header.ch[ch-1].unit_actual)
        self.coupling = w.header.ch[ch-1].coupling.name.upper()
        self.y_scale = -self.volt_scale
        self.y_offset = self.volt_offset

        if self.enabled:
            if ch == 1:
                self.raw = np.array(w.header.raw_1, dtype=np.uint8)

            if ch == 2:
                self.raw = np.array(w.header.raw_2, dtype=np.uint8)

            if ch == 3:
                self.raw = np.array(w.header.raw_3, dtype=np.uint8)

            if ch == 4:
                self.raw = np.array(w.header.raw_4, dtype=np.uint8)

        self.calc_times_and_volts()


    def ds4000(self, w, ch):
        """Interpret waveform for the Rigol DS4000 series."""
        self.time_offset = w.header.time_offset
        self.time_scale = w.header.time_scale
        self.points = w.header.points
        self.firmware = w.header.firmware_version
        self.coupling = w.header.ch[ch-1].coupling.name.upper()
        self.y_scale = -self.volt_scale
        self.y_offset = self.volt_offset

        if self.enabled:
            if ch == 1:
                self.raw = np.array(w.header.raw_1, dtype=np.uint8)

            elif ch == 2:
                self.raw = np.array(w.header.raw_2, dtype=np.uint8)

            elif ch == 3:
                self.raw = np.array(w.header.raw_3, dtype=np.uint8)

            elif ch == 4:
                self.raw = np.array(w.header.raw_4, dtype=np.uint8)

        self.calc_times_and_volts()


    def ds6000(self, w, ch):
        """Interpret waveform for the Rigol DS6000 series."""
        self.time_offset = w.header.time_offset
        self.time_scale = w.header.time_scale
        self.points = w.header.points
        self.firmware = w.header.firmware_version
        self.coupling = w.header.ch[ch-1].coupling.name.upper()
        self.unit = w.header.ch[ch-1].unit

        if self.enabled:
            if ch == 1:
                self.raw = np.array(w.header.raw_1, dtype=np.uint8)

            if ch == 2:
                self.raw = np.array(w.header.raw_2, dtype=np.uint8)

            if ch == 3:
                self.raw = np.array(w.header.raw_3, dtype=np.uint8)

            if ch == 4:
                self.raw = np.array(w.header.raw_4, dtype=np.uint8)

        self.calc_times_and_volts()



#pylint: disable=invalid-name
#pylint: disable=too-many-instance-attributes
"""
Extract signals or description from Rigol 1000E Oscilloscope waveform file.

Use like this::

    import RigolWFM.wfm as rigol

    waveform = rigol.Wfm.from_file("file_name.wfm", 'E')
    description = waveform.describe()
    print(description)

"""
import tempfile
import os.path
import wave
import urllib.parse

import requests
import matplotlib.pyplot as plt

import RigolWFM.wfm1000c
import RigolWFM.wfm1000e
import RigolWFM.wfm1000z
import RigolWFM.wfm2000
import RigolWFM.wfm4000
import RigolWFM.wfm6000
import RigolWFM.channel

# not implemented
# DS1000B_scopes = ["B", "1000B", "DS1000B",
#                  "DS1074B", "DS1104B", "DS1204B"]

# untested
DS1000C_scopes = ["C", "1000C", "DS1000C",
                  "DS1000CD", "DS1000C", "DS1000MD", "DS1000M",
                  "DS1302CA", "DS1202CA", "DS1102CA", "DS1062CA"]

# tested
DS1000E_scopes = ["E", "1000E", "DS1000E",
                  "D", "1000D", "DS1000D",
                  "DS1102E", "DS1052E", "DS1102D", "DS1052D"]

# tested, wonky voltages
DS1000Z_scopes = ["Z", "1000Z", "DS1000Z",
                  "DS1202Z",
                  "DS1054Z", "MSO1054Z",
                  "DS1074Z", "MSO1074Z", "DS1074Z-S",
                  "DS1104Z", "MSO1104Z", "DS1104Z-S"]

# tested
DS2000_scopes = ["2", "2000", "DS2000", "DS2072A",
                 "DS2102A", "MSO2102A", "MSO2102A-S",
                 "DS2202A", "MSO2202A", "MSO2202A-S",
                 "DS2302A", "MSO2302A", "MSO2302A-S"]

# tested
DS4000_scopes = ["4", "4000", "DS4000",
                 "DS4054", "DS4052", "DS4034", "DS4032", "DS4024",
                 "DS4022", "DS4014", "DS4012", "MSO4054", "MSO4052", "MSO4034",
                 "MSO4032", "MSO4024", "MSO4022", "MSO4014", "MSO4012"]

# untested
DS6000_scopes = ["6", "6000", "DS6000",
                 "DS6062", "DS6064", "DS6102", "DS6104"]


def valid_scope_list():
    """List all the oscilloscope types."""
    s = "\nValid types are:\n"
    s += ", ".join(DS1000C_scopes) + "\n"
    s += ", ".join(DS1000E_scopes) + "\n"
    s += ", ".join(DS1000Z_scopes) + "\n"
    s += ", ".join(DS2000_scopes) + "\n"
    s += ", ".join(DS4000_scopes) + "\n"
    s += ", ".join(DS6000_scopes) + "\n"
    return s


class Read_WFM_Error(Exception):
    """Generic Read Error."""


class Parse_WFM_Error(Exception):
    """Generic Parse Error."""


class Invalid_URL(Exception):
    """Cannot use this URL.  It must start with https or http."""


class Unknown_Scope_Error(Exception):
    """Not one of the listed Rigol oscilloscopes."""


class Write_WAV_Error(Exception):
    """Something went wrong while writing the .wave file."""


class Channel_Not_In_WFM_Error(Exception):
    """The channel is not in the .wfm file."""


class Wfm():
    """Class with parsed data from a .wfm file."""

    def __init__(self, file_name):
        """Initialize a Wfm object from a file."""
        self.channels = []
        self.original_name = file_name
        self.file_name = file_name
        self.basename = file_name
        self.firmware = 'unknown'

        # there are multiple possible scope names
        # 1. user_name   - the name passed to the program
        # 2. header_name - the name found in the header of the file
        # 3. parser_name - the name of parser used
        self.user_name = 'unknown'
        self.parser_name = 'unknown'
        self.header_name = 'unknown'

    @classmethod
    def from_file(cls, file_name, model):
        """Create Wfm object from a file."""
        # ensure that file exists
        try:
            f = open(file_name, 'rb')
            f.close()
        except IOError as e:
            raise Read_WFM_Error(e)

        new_wfm = cls(file_name)
        new_wfm.original_name = file_name
        new_wfm.file_name = file_name
        new_wfm.basename = os.path.basename(file_name)

        # parse the waveform
        umodel = model.upper()

        if umodel in DS1000C_scopes:
            w = RigolWFM.wfm1000c.Wfm1000c.from_file(file_name)
            new_wfm.header_name = 'DS1000C'

        elif umodel in DS1000E_scopes:
            w = RigolWFM.wfm1000e.Wfm1000e.from_file(file_name)
            new_wfm.header_name = 'DS1000E'

        elif umodel in DS1000Z_scopes:
            w = RigolWFM.wfm1000z.Wfm1000z.from_file(file_name)
            new_wfm.header_name = w.preheader.model_number

        elif umodel in DS2000_scopes:
            w = RigolWFM.wfm2000.Wfm2000.from_file(file_name)
            new_wfm.header_name = 'DS2000'

        elif umodel in DS4000_scopes:
            w = RigolWFM.wfm4000.Wfm4000.from_file(file_name)
            new_wfm.header_name = w.header.model_number

        elif umodel in DS6000_scopes:
            w = RigolWFM.wfm6000.Wfm6000.from_file(file_name)
            new_wfm.header_name = w.header.model_number

        else:
            print("Unknown Rigol oscilloscope type: '%s'" % umodel)
            print(valid_scope_list())
            return new_wfm

        new_wfm.user_name = model
        pname = str(w).split(".")[1]
        new_wfm.parser_name = pname


        # assemble into uniform set of names
        enabled_channels = 0
        for ch_number in range(1, 5):

            ch = RigolWFM.channel.Channel(w, ch_number, pname, enabled_channels)
            if ch.enabled:
                new_wfm.channels.append(ch)
                enabled_channels += 1

        new_wfm.firmware = new_wfm.channels[0].firmware
        return new_wfm

    @classmethod
    def from_url(cls, url, model):
        """
        Return a waveform object given a URL.

        This is a bit complicated because the parser must have a local file
        to work with.  The process is to download the file to a temporary
        location and then process that file.  There is a lot that can go
        wrong - bad url, bad download, or an error parsing the file.
        """
        u = urllib.parse.urlparse(url)
        scheme = u[0]

        if scheme not in ['http', 'https']:
            raise Invalid_URL()

        try:
            # need a local file for conversion, download url and save as tempfile
            print("downloading '%s'" % url)
            r = requests.get(url, allow_redirects=True)
            if not r.ok:
                error_string = "Downloading URL '%s' failed: '%s'" % (
                    url, r.reason)
                raise Read_WFM_Error(error_string)

            f = tempfile.NamedTemporaryFile()
            f.write(r.content)
            f.seek(0)
            working_name = f.name

            try:
                new_wfm = cls.from_file(working_name, model)
                new_wfm.original_name = url
                # extract the simple name
                rawpath = u[2]
                path = urllib.parse.unquote(rawpath)
                new_wfm.basename = os.path.basename(path)
                return new_wfm
            except Exception as e:
                raise Parse_WFM_Error(e)

        except Exception as e:
            raise Parse_WFM_Error(e)

    def describe(self):
        """Return a string describing the contents of a Rigol wfm file."""
        s = "    General:\n"
        s += '        File Model   = %s\n' % self.parser_name
        s += "        User Model   = %s\n" % self.user_name
        s += '        Parser Model = %s\n' % self.parser_name
        s += "        Firmware     = %s\n" % self.firmware
        s += '        Filename     = %s\n' % self.basename
        s += '        Channels     = ['

        first = True
        for ch in self.channels:
            if not first:
                s += ', '
            s += '%s' % ch.channel_number
            first = False
        s += ']\n\n'

        for ch in self.channels:
            s += str(ch)
            s += "\n"
        return s

    def best_scaling(self):
        """Return appropriate scaling for plot."""
        v_scale = 1e-12
        v_prefix = ''
        h_scale = 1e-12
        h_prefix = ''
        for ch in self.channels:
            v, p = RigolWFM.channel.best_scale(ch.volt_per_division)
            if v > v_scale:
                v_scale = v
                v_prefix = p
            h, p = RigolWFM.channel.best_scale(ch.time_scale)
            if h > h_scale:
                h_scale = h
                h_prefix = p
        return h_scale, h_prefix, v_scale, v_prefix

    def plot(self):
        """Plot the data."""
        h_scale, h_prefix, v_scale, v_prefix = self.best_scaling()
        colors = ['red', 'blue', 'orange', 'magenta']

        for i, ch in enumerate(self.channels):
            plt.plot(ch.times*h_scale, ch.volts*v_scale,
                     label=ch.name, color=colors[i])

        plt.xlabel("Time (%ss)" % h_prefix)
        plt.ylabel("Voltage (%sV)" % v_prefix)
        plt.title(self.basename)
        plt.legend(loc='upper right')

    def csv(self):
        """Return a string of comma separated values."""
        if len(self.channels) == 0:
            return ''

        h_scale, h_prefix, v_scale, v_prefix = self.best_scaling()

        s = 'X'
        for ch in self.channels:
            s += ",%s" % ch.name
        s += ",Start,Increment\n"

        # just output the display 100 pts/division
        ch = self.channels[0]
        incr = ch.time_scale/100
        off = -6 * ch.time_scale
        s += '%ss' % h_prefix
        for ch in self.channels:
            s += ",%s%s" % (v_prefix, ch.unit.name.upper())
        s += ",%e,%e\n" % (off, incr)

        for i in range(self.channels[0].points):
            s += "%.6f" % (ch.times[i]*h_scale)
            for ch in self.channels:
                s += ",%.2f" % (ch.volts[i]*v_scale)
            s += "\n"
        return s

    def wav(self, wav_filename, channel=1):
        """Save data as a WAV file for use with LTSpice."""
        c = None
        for ch in self.channels:
            if channel == ch.channel_number:
                c = ch
                break

        if c is None:
            raise Channel_Not_In_WFM_Error()

        sample_rate = 1.0/c.seconds_per_point

        wavef = wave.open(wav_filename, 'wb')
        wavef.setnchannels(len(self.channels))  # mono
        wavef.setsampwidth(1)
        wavef.setframerate(sample_rate)
        wavef.setcomptype('NONE', '')
        wavef.setnframes(self.channels[0].points)

        for ch in self.channels:
            wavef.writeframes(ch.raw)
        wavef.close()



#pylint: disable=invalid-name
#pylint: disable=missing-function-docstring
#pylint: disable=unused-argument
"""
Command line utility to convert Rigol .wfm files.

Use like this::

    wfmconvert 1102E info DS1102E-A.wfm
    wfmconvert 1102E csv DS1102E-A.wfm
    wfmconvert 1102E wav DS1102E-A.wfm
"""

import os
import sys
import argparse

import RigolWFM.wfm as rigol

def info(args, scope_data, infile):
    """Create a string that describes content of .wfm file."""
    s = scope_data.describe()
    print(s)

def csv(args, scope_data, infile):
    """Create a file with comma separated values."""
    csv_name = os.path.splitext(infile)[0] + '.csv'

    if os.path.isfile(csv_name) and not args.force:
        print("'%s' exists, use --force to overwrite" % csv_name)
        return

    s = scope_data.csv()
    with open(csv_name, 'wb') as f:
        b = s.encode(encoding='utf-8')
        f.write(b)

def wav(args, scope_data, infile):
    """Create an audible .wav file for use in LTSpice."""
    wav_name = os.path.splitext(infile)[0] + '.wav'
    if os.path.isfile(wav_name) and not args.force:
        print("'%s' exists, use --force to overwrite" % wav_name)
        return

    scope_data.wav(wav_name, channel=args.channel)

def main():
    """Parse console command line arguments."""
    parser = argparse.ArgumentParser(
        prog='wfmconvert',
        description='Parse Rigol WFM files.',
        formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        '--force',
        action='store_true',
        help="overwrite existing csv or wav files"
    )

    parser.add_argument(
        '--channel',
        type=int,
        default=1,
        help="designate channel to save in .wav file"
    )

    parser.add_argument(
        'model',
        type=str,
        help='the type of scope that created the WFM file' + rigol.valid_scope_list()
    )

    parser.add_argument(
        dest='action',
        choices=['csv', 'info', 'wav'],
        help="Action to perform on the WFM file"
    )

    parser.add_argument(
        'infile',
        type=str,
        nargs='+',
        help="Input WFM file"
    )

    args = parser.parse_args()

    actionMap = {"info": info, "csv": csv, "wav": wav}

    for filename in args.infile:
        try:
            scope_data = rigol.Wfm.from_file(filename, args.model)
            actionMap[args.action](args, scope_data, filename)

        except rigol.Parse_WFM_Error as e:
            print("Format does not follow a known format.", file=sys.stderr)
            print("To help development, post report this error\n", file=sys.stderr)
            print("as an issue to https://github.com/scottprahl/RigolWFM\n", file=sys.stderr)
            print(e, file=sys.stderr)
            sys.exit()

if __name__ == "__main__":
    main()



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
from kaitaistruct import __version__ as ks_version, KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum


if parse_version(ks_version) < parse_version('0.7'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.7 or later is required, but you have %s" % (ks_version))

class Wfm2000(KaitaiStruct):

    class UnitEnum(Enum):
        w = 0
        a = 1
        v = 2
        u = 3

    class ProbeRatioEnum(Enum):
        x0_01 = 0
        x0_02 = 1
        x0_05 = 2
        x0_1 = 3
        x0_2 = 4
        x0_5 = 5
        x1 = 6
        x2 = 7
        x5 = 8
        x10 = 9
        x20 = 10
        x50 = 11
        x100 = 12
        x200 = 13
        x500 = 14
        x1000 = 15

    class ImpedanceEnum(Enum):
        ohm_50 = 0
        ohm_1meg = 1

    class ProbeTypeEnum(Enum):
        normal_type = 0
        differential = 1

    class ProbeEnum(Enum):
        single = 0
        diff = 1

    class FilterEnum(Enum):
        low_pass = 0
        high_pass = 1
        band_pass = 2
        band_reject = 3

    class BandwidthEnum(Enum):
        no_limit = 0
        mhz_20 = 1
        mhz_100 = 2
        mhz_200 = 3
        mhz_250 = 4

    class TimeEnum(Enum):
        yt = 0
        xy = 1
        roll = 2

    class AcquisitionEnum(Enum):
        normal = 0
        average = 1
        peak = 2
        high_resolution = 3

    class CouplingEnum(Enum):
        dc = 0
        ac = 1
        gnd = 2
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        pass

    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.magic = self._io.ensure_fixed_contents(b"\xA5\xA5\x38\x00")
            self.model_number = (KaitaiStream.bytes_terminate(self._io.read_bytes(20), 0, False)).decode(u"ascii")
            self.firmware_version = (KaitaiStream.bytes_terminate(self._io.read_bytes(20), 0, False)).decode(u"ascii")
            self.block_num = self._io.ensure_fixed_contents(b"\x01\x00")
            self.file_version = self._io.read_u2le()
            self.unused_1 = self._io.read_bytes(8)
            self.crc = self._io.read_u4le()
            self.structure_size = self._io.read_u2le()
            self.structure_version = self._io.read_u2le()
            self.enabled = self._root.ChannelMask(self._io, self, self._root)
            self.extra_1a = self._io.read_bytes(2)
            self.channel_offset = [None] * (4)
            for i in range(4):
                self.channel_offset[i] = self._io.read_u4le()

            self.acquisition_mode = self._root.AcquisitionEnum(self._io.read_u2le())
            self.average_time = self._io.read_u2le()
            self.sample_mode = self._io.read_u2le()
            self.extra_1b = self._io.read_bytes(2)
            self.mem_depth = self._io.read_u4le()
            self.sample_rate_hz = self._io.read_f4le()
            self.extra_1c = self._io.read_bytes(2)
            self.time_mode = self._root.TimeEnum(self._io.read_u2le())
            self.time_scale_ps = self._io.read_u8le()
            self.time_offset_ps = self._io.read_s8le()
            self.ch = [None] * (4)
            for i in range(4):
                self.ch[i] = self._root.ChannelHeader(self._io, self, self._root)

            self.setup_size = self._io.read_u4le()
            self.setup_offset = self._io.read_u4le()
            self.wfm_offset = self._io.read_u4le()
            self.storage_depth = self._io.read_u4le()
            self.z_pt_offset = self._io.read_u4le()
            self.wfm_len = self._io.read_u4le()
            self.equ_coarse = [None] * (2)
            for i in range(2):
                self.equ_coarse[i] = self._io.read_u2le()

            self.equ_fine = [None] * (2)
            for i in range(2):
                self.equ_fine[i] = self._io.read_u2le()

            self.mem_start_addr = [None] * (2)
            for i in range(2):
                self.mem_start_addr[i] = self._io.read_u4le()

            self.mem_last_addr = self._io.read_u4le()
            self.mem_length = self._io.read_u4le()
            self.channel_depth = self._io.read_u4le()
            self.bank_size = self._io.read_u4le()
            self.bank_offset = self._io.read_u4le()
            self.real_offset = self._io.read_u4le()
            self.real_ch_offset = self._io.read_u4le()
            self.horiz_slow_force_stop_frame_boolean = self._io.read_u1()
            self.get_spu_dig_data_status_boolean = self._io.read_u1()
            self.spu_load_data_status_boolean = self._io.read_u1()
            self.trig_delay_mem_offset = self._io.read_s8le()
            self.trig_delay_view_offset = self._io.read_s8le()
            self.mem_offset_compensate = self._io.read_s8le()
            self.slow_delta_wave_length = self._io.read_s8le()
            self.channel_pos_max_delay = self._io.read_s8le()
            self.channel_neg_min_delay = self._io.read_s8le()
            self.real_sa_dot_period = self._io.read_u8le()
            self.mem_offset_base = self._io.read_u8le()
            self.trig_type_delta_delay = self._io.read_s4le()
            self.ch1_delay = self._io.read_s4le()
            self.ch2_delay = self._io.read_s4le()
            self.channel_delay_to_mem_len = self._io.read_u4le()
            self.spu_mem_bank_size = self._io.read_u4le()
            self.roll_scrn_wave_length = self._io.read_u2le()
            self.trigger_source = self._io.read_u1()
            self.cal_index = self._io.read_u1()
            self.record_frame_index = self._io.read_u4le()
            self.frame_cur = self._io.read_u4le()

        @property
        def points(self):
            if hasattr(self, '_m_points'):
                return self._m_points if hasattr(self, '_m_points') else None

            self._m_points = self._root.header.mem_depth
            return self._m_points if hasattr(self, '_m_points') else None

        @property
        def time_offset(self):
            if hasattr(self, '_m_time_offset'):
                return self._m_time_offset if hasattr(self, '_m_time_offset') else None

            self._m_time_offset = (1.0E-12 * self.time_offset_ps)
            return self._m_time_offset if hasattr(self, '_m_time_offset') else None

        @property
        def raw_2(self):
            if hasattr(self, '_m_raw_2'):
                return self._m_raw_2 if hasattr(self, '_m_raw_2') else None

            if self.channel_offset[1] > 0:
                io = self._root._io
                _pos = io.pos()
                io.seek(self.channel_offset[1])
                self._m_raw_2 = [None] * (self.storage_depth)
                for i in range(self.storage_depth):
                    self._m_raw_2[i] = io.read_u1()

                io.seek(_pos)

            return self._m_raw_2 if hasattr(self, '_m_raw_2') else None

        @property
        def time_scale(self):
            if hasattr(self, '_m_time_scale'):
                return self._m_time_scale if hasattr(self, '_m_time_scale') else None

            self._m_time_scale = (1.0E-12 * self.time_scale_ps)
            return self._m_time_scale if hasattr(self, '_m_time_scale') else None

        @property
        def seconds_per_point(self):
            if hasattr(self, '_m_seconds_per_point'):
                return self._m_seconds_per_point if hasattr(self, '_m_seconds_per_point') else None

            self._m_seconds_per_point = (1 / self.sample_rate_hz)
            return self._m_seconds_per_point if hasattr(self, '_m_seconds_per_point') else None

        @property
        def raw_4(self):
            if hasattr(self, '_m_raw_4'):
                return self._m_raw_4 if hasattr(self, '_m_raw_4') else None

            if self.channel_offset[3] > 0:
                io = self._root._io
                _pos = io.pos()
                io.seek(self.channel_offset[3])
                self._m_raw_4 = [None] * (self.storage_depth)
                for i in range(self.storage_depth):
                    self._m_raw_4[i] = io.read_u1()

                io.seek(_pos)

            return self._m_raw_4 if hasattr(self, '_m_raw_4') else None

        @property
        def raw_3(self):
            if hasattr(self, '_m_raw_3'):
                return self._m_raw_3 if hasattr(self, '_m_raw_3') else None

            if self.channel_offset[2] > 0:
                io = self._root._io
                _pos = io.pos()
                io.seek(self.channel_offset[2])
                self._m_raw_3 = [None] * (self.storage_depth)
                for i in range(self.storage_depth):
                    self._m_raw_3[i] = io.read_u1()

                io.seek(_pos)

            return self._m_raw_3 if hasattr(self, '_m_raw_3') else None

        @property
        def raw_1(self):
            if hasattr(self, '_m_raw_1'):
                return self._m_raw_1 if hasattr(self, '_m_raw_1') else None

            if self.channel_offset[0] > 0:
                io = self._root._io
                _pos = io.pos()
                io.seek(self.channel_offset[0])
                self._m_raw_1 = [None] * (self.storage_depth)
                for i in range(self.storage_depth):
                    self._m_raw_1[i] = io.read_u1()

                io.seek(_pos)

            return self._m_raw_1 if hasattr(self, '_m_raw_1') else None


    class ChannelHeader(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.enabled_temp = self._io.read_u1()
            self.coupling = self._root.CouplingEnum(self._io.read_bits_int(2))
            self.skip_coupling = self._io.read_bits_int(6)
            self._io.align_to_byte()
            self.bandwidth_limit = self._root.BandwidthEnum(self._io.read_u1())
            self.probe_type = self._root.ProbeTypeEnum(self._io.read_u1())
            self.probe_ratio = self._root.ProbeRatioEnum(self._io.read_u1())
            self.probe_diff = self._root.ProbeEnum(self._io.read_u1())
            self.probe_signal = self._root.ProbeEnum(self._io.read_u1())
            self.probe_impedance = self._root.ImpedanceEnum(self._io.read_u1())
            self.volt_per_division = self._io.read_f4le()
            self.volt_offset = self._io.read_f4le()
            self.inverted_temp = self._io.read_u1()
            self.unit_temp = self._io.read_u1()
            self.filter_enabled = self._io.read_u1()
            self.filter_type = self._io.read_u1()
            self.filter_high = self._io.read_u4le()
            self.filter_low = self._io.read_u4le()

        @property
        def unit(self):
            if hasattr(self, '_m_unit'):
                return self._m_unit if hasattr(self, '_m_unit') else None

            self._m_unit = self.unit_actual
            return self._m_unit if hasattr(self, '_m_unit') else None

        @property
        def unit_actual(self):
            if hasattr(self, '_m_unit_actual'):
                return self._m_unit_actual if hasattr(self, '_m_unit_actual') else None

            self._m_unit_actual = (self.unit_temp if self.enabled_temp == 1 else self.inverted_temp)
            return self._m_unit_actual if hasattr(self, '_m_unit_actual') else None

        @property
        def inverted(self):
            if hasattr(self, '_m_inverted'):
                return self._m_inverted if hasattr(self, '_m_inverted') else None

            self._m_inverted = (True if self.inverted_actual == 1 else False)
            return self._m_inverted if hasattr(self, '_m_inverted') else None

        @property
        def volt_scale(self):
            if hasattr(self, '_m_volt_scale'):
                return self._m_volt_scale if hasattr(self, '_m_volt_scale') else None

            self._m_volt_scale = (self.volt_signed / 25.0)
            return self._m_volt_scale if hasattr(self, '_m_volt_scale') else None

        @property
        def inverted_actual(self):
            if hasattr(self, '_m_inverted_actual'):
                return self._m_inverted_actual if hasattr(self, '_m_inverted_actual') else None

            self._m_inverted_actual = (self.inverted_temp if self.enabled_temp == 1 else self.unit_temp)
            return self._m_inverted_actual if hasattr(self, '_m_inverted_actual') else None

        @property
        def probe_value(self):
            if hasattr(self, '_m_probe_value'):
                return self._m_probe_value if hasattr(self, '_m_probe_value') else None

            self._m_probe_value = (0.01 if self.probe_ratio == self._root.ProbeRatioEnum.x0_01 else (0.02 if self.probe_ratio == self._root.ProbeRatioEnum.x0_02 else (0.05 if self.probe_ratio == self._root.ProbeRatioEnum.x0_05 else (0.1 if self.probe_ratio == self._root.ProbeRatioEnum.x0_1 else (0.2 if self.probe_ratio == self._root.ProbeRatioEnum.x0_2 else (0.5 if self.probe_ratio == self._root.ProbeRatioEnum.x0_5 else (1.0 if self.probe_ratio == self._root.ProbeRatioEnum.x1 else (2.0 if self.probe_ratio == self._root.ProbeRatioEnum.x2 else (5.0 if self.probe_ratio == self._root.ProbeRatioEnum.x5 else (10.0 if self.probe_ratio == self._root.ProbeRatioEnum.x10 else (20.0 if self.probe_ratio == self._root.ProbeRatioEnum.x20 else (50.0 if self.probe_ratio == self._root.ProbeRatioEnum.x50 else (100.0 if self.probe_ratio == self._root.ProbeRatioEnum.x100 else (200.0 if self.probe_ratio == self._root.ProbeRatioEnum.x200 else (500.0 if self.probe_ratio == self._root.ProbeRatioEnum.x500 else 1000.0)))))))))))))))
            return self._m_probe_value if hasattr(self, '_m_probe_value') else None

        @property
        def enabled(self):
            if hasattr(self, '_m_enabled'):
                return self._m_enabled if hasattr(self, '_m_enabled') else None

            self._m_enabled = (True if self.enabled_temp != 0 else False)
            return self._m_enabled if hasattr(self, '_m_enabled') else None

        @property
        def volt_signed(self):
            if hasattr(self, '_m_volt_signed'):
                return self._m_volt_signed if hasattr(self, '_m_volt_signed') else None

            self._m_volt_signed = ((-1.0 * self.volt_per_division) if self.inverted else (1.0 * self.volt_per_division))
            return self._m_volt_signed if hasattr(self, '_m_volt_signed') else None


    class ChannelMask(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.unused_1 = self._io.read_bits_int(4)
            self.channel_4 = self._io.read_bits_int(1) != 0
            self.channel_3 = self._io.read_bits_int(1) != 0
            self.channel_2 = self._io.read_bits_int(1) != 0
            self.channel_1 = self._io.read_bits_int(1) != 0
            self.unused_2 = self._io.read_bits_int(7)
            self.interwoven = self._io.read_bits_int(1) != 0


    @property
    def header(self):
        if hasattr(self, '_m_header'):
            return self._m_header if hasattr(self, '_m_header') else None

        _pos = self._io.pos()
        self._io.seek(0)
        self._m_header = self._root.Header(self._io, self, self._root)
        self._io.seek(_pos)
        return self._m_header if hasattr(self, '_m_header') else None


